/*
 * File:        TFT_keypad_BRL4.c
 * Author:      Bruce Land
 * Adapted from:
 *              main.c by
 * Author:      Syed Tahmid Mahbub
 * Target PIC:  PIC32MX250F128B
 */

// graphics libraries
#include "config.h"
#include "tft_master.h"
#include "tft_gfx.h"
#include <stdlib.h> // need for rand function
#include <string.h>

// threading library
#include <plib.h>
#include <math.h> // for sine
#include "pt_cornell_1_2_1.h"

#define	SYS_FREQ 60000000

// UART parameters
#define BAUDRATE 9600 // must match PC end
#define PB_DIVISOR (1 << OSCCONbits.PBDIV) // read the peripheral bus divider, FPBDIV
#define PB_FREQ SYS_FREQ/PB_DIVISOR // periperhal bus frequency
#define clrscr() printf( "\x1b[2J")
#define home()   printf( "\x1b[H")
#define pcr()    printf( '\r')
#define crlf     putchar(0x0a); putchar(0x0d);
#define backspace 0x7f // make sure your backspace matches this!
#define max_chars 32 // for input/output buffer

// Fix16 Data Type declarations and common constants 
// ============================================
typedef signed int fix16;
#define multfix16(a,b) ((fix16)(((( signed long long)(a))*(( signed long long)(b)))>>16)) //multiply two fixed 16:16
#define float2fix16(a) ((fix16)((a)*65536.0)) // 2^16
#define fix2float16(a) ((float)(a)/65536.0)
#define fix2int16(a)   ((int)((a)>>16))
#define int2fix16(a)   ((fix16)((a)<<16))
#define divfix16(a,b)  ((fix16)((((signed long long)(a)<<16)/(b)))) 
#define sqrtfix16(a)   (float2fix16(sqrt(fix2float16(a)))) 
#define absfix16(a)    abs(a)
#define onefix16       0x00010000 // int2fix16(1)

#define fix16_60000 int2fix16(60000)
#define clock_tick_ms_float 256.0/SYS_FREQ
#define clock_tick_ms float2fix16(clock_tick_ms_float)

#define MIN_PWM_PERIOD_RIGHT 3925
#define MIN_PWM_PERIOD_LEFT 3825
// ============================================

// string buffer
char buffer[60];

// Enum used for referencing direction of movement
enum MOVE_DIRECTION
{
    move_stop,
    move_left,
    move_right,
    move_straight,
    move_back
};

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_servo, pt_color, pt_anim, pt_serial, pt_timekeeper;

// UART pt inits
static struct pt pt_input, pt_output, pt_DMA_output;

#define CW   1
#define CCW -1
#define DEFAULT_STEPPER_THREAD_YIELD_TIME 100

volatile int rotation_amount = 0;
volatile int servo_enabled = 0;
volatile int rotation_direction;
volatile int stepper_position = 24;
volatile int stepper_movement_time = 0;
volatile int stepper_thread_yield_time = DEFAULT_STEPPER_THREAD_YIELD_TIME;

void set_stepper_position(int angle){
    if ( servo_enabled == 1){ return; } // abort if stepper is already executing
    
    // convert angle to stepper position
    int angle_convert = ((float)angle) / 7.5;
    
    int relative_angle = angle_convert - stepper_position;
    if (relative_angle < 0){
        relative_angle = relative_angle * -1;
        rotation_direction = CW;
    } else {
        rotation_direction = CCW;
    }
    
    // Set stepper thread yield time
    if (stepper_movement_time > 0){
        stepper_thread_yield_time = stepper_movement_time / relative_angle;
    } else {
        stepper_thread_yield_time = DEFAULT_STEPPER_THREAD_YIELD_TIME;
    }
    
    stepper_position = angle_convert;
    rotation_amount = relative_angle;
    servo_enabled = 1;
}

// NOT USED
#define ROBOT_TURN_SPEED 50
#define ROBOT_TURN_TIME  10
volatile int stop_time; 
volatile int robot_turning = 0;

// default value
volatile int motor_speed = 30;

unsigned int sys_time_seconds;

static PT_THREAD(protothread_timekeeper(struct pt *pt)) {
    PT_BEGIN(pt);

    while (1) {
        PT_YIELD_TIME_msec(1000);
        sys_time_seconds++;
        if (robot_turning && sys_time_seconds >= stop_time){ // NOT USED
            robot_turning = 0;
            move_robot(move_stop, motor_speed);
        }
    }
    PT_END(pt);
}

#define STEPPER_OUT1_SET        mPORTBSetBits(BIT_9)
#define STEPPER_OUT1_CLEAR      mPORTBClearBits(BIT_9)

#define STEPPER_OUT2_SET        mPORTBSetBits(BIT_3);
#define STEPPER_OUT2_CLEAR      mPORTBClearBits(BIT_3);

#define STEPPER_OUT3_SET        mPORTBSetBits(BIT_13);
#define STEPPER_OUT3_CLEAR      mPORTBClearBits(BIT_13);

#define STEPPER_OUT4_SET        mPORTBSetBits(BIT_15);
#define STEPPER_OUT4_CLEAR      mPORTBClearBits(BIT_15);

// Helper function for moving the servo motors
// dir - direction of enum type MOVE_DIRECTION
// speed - number between 0 (no movement) and 100 (full speed)
// Note: if dir = move_stop, the speed parameter is irrelevant
void move_robot(enum MOVE_DIRECTION dir, int speed){
    
    // Default configurations for right wheel (default backward)
    /*OpenOC1(OC_ON | OC_TIMER3_SRC | OC_PWM_FAULT_PIN_ENABLE, generate_period >> 1, generate_period >> 2); //OC_PWM_FAULT_PIN_ENABLE
    PPSOutput(1, RPA0, OC1);
    SetDCOC1PWM(3825);  
    mPORTBSetPinsDigitalOut(BIT_7);
    mPORTBClearBits (BIT_7);
    
        
    // Default configurations for left wheel (default forward)
    OpenOC2(OC_ON | OC_TIMER3_SRC | OC_PWM_FAULT_PIN_ENABLE, generate_period >> 1, generate_period >> 2); //OC_PWM_FAULT_PIN_ENABLE
    PPSOutput(2, RPB8, OC2);
    SetDCOC2PWM(3825);
    mPORTBSetPinsDigitalOut(BIT_5);
    mPORTBClearBits (BIT_5);*/
    
    int pwm_period_left  = (((10000-MIN_PWM_PERIOD_LEFT)*(speed))/100) + MIN_PWM_PERIOD_LEFT;
    int pwm_period_right = (((10000-MIN_PWM_PERIOD_RIGHT)*(speed))/100) + MIN_PWM_PERIOD_RIGHT;
    
    if (dir == move_stop) {
        // Right wheel
        SetDCOC1PWM(0);
        
        // Left wheel
        SetDCOC2PWM(0);
    }
    else if (dir == move_left) {
        // Right wheel
        PPSOutput(1, RPB7, OC1);
        SetDCOC1PWM(pwm_period_right);  
        PPSOutput(1, RPA0, NULL);
        mPORTASetPinsDigitalOut(BIT_0);
        mPORTAClearBits (BIT_0);
        
        // Left wheel
        PPSOutput(2, RPB8, OC2);
        SetDCOC2PWM(pwm_period_left);
        PPSOutput(2, RPB5, NULL);
        mPORTBSetPinsDigitalOut(BIT_5);
        mPORTBClearBits (BIT_5);
    }
    else if (dir == move_right) {
        // Left wheel
        PPSOutput(2, RPB5, OC2);
        SetDCOC2PWM(pwm_period_left);
        PPSOutput(2, RPB8, NULL);
        mPORTBSetPinsDigitalOut(BIT_8);
        mPORTBClearBits (BIT_8);
        
        // Right wheel
        PPSOutput(1, RPA0, OC1);
        SetDCOC1PWM(pwm_period_right);  
        PPSOutput(1, RPB7, NULL);
        mPORTBSetPinsDigitalOut(BIT_7);
        mPORTBClearBits (BIT_7);    
    }
    else if (dir == move_straight) {
        // Right wheel
        PPSOutput(1, RPB7, OC1);
        SetDCOC1PWM(pwm_period_right);  
        PPSOutput(1, RPA0, NULL);
        mPORTASetPinsDigitalOut(BIT_0);
        mPORTAClearBits (BIT_0);
        

        
                // Left wheel
        PPSOutput(2, RPB5, OC2);
        SetDCOC2PWM(pwm_period_left);
        PPSOutput(2, RPB8, NULL);
        mPORTBSetPinsDigitalOut(BIT_8);
        mPORTBClearBits (BIT_8);
    }
    
    else if (dir == move_back) {
        // Right wheel
        PPSOutput(1, RPA0, OC1);
        SetDCOC1PWM(pwm_period_right);  
        PPSOutput(1, RPB7, NULL);
        mPORTBSetPinsDigitalOut(BIT_7);
        mPORTBClearBits (BIT_7);
        
        // Left wheel
        PPSOutput(2, RPB8, OC2);
        SetDCOC2PWM(pwm_period_left);
        PPSOutput(2, RPB5, NULL);
        mPORTBSetPinsDigitalOut(BIT_5);
        mPORTBClearBits (BIT_5);
    }
    
    
    /**//*
    //Open timer 3
    OpenTimer3(T3_ON | T3_SOURCE_INT | T3_PS_1_1, generate_period);
    OpenOC1(OC_ON | OC_TIMER3_SRC | OC_PWM_FAULT_PIN_ENABLE, generate_period >> 1, generate_period >> 2); //OC_PWM_FAULT_PIN_ENABLE
    PPSOutput(1, RPA0, OC1);
    SetDCOC1PWM(3825);
    
    mPORTBSetPinsDigitalOut(BIT_7);
    mPORTBSetBits (BIT_7);

    int count = 0;
    while (count < 65500000) {
        count ++;
    }
    PPSOutput(1, RPB7, OC1);
    PPSOutput(1, RPA0, NULL);
    
    mPORTASetPinsDigitalOut(BIT_0);
    mPORTASetBits(BIT_0);
    
    // TEST
    count = 0;
    //OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, generate_period);
    OpenOC2(OC_ON | OC_TIMER3_SRC | OC_PWM_FAULT_PIN_ENABLE, generate_period >> 1, generate_period >> 2); //OC_PWM_FAULT_PIN_ENABLE
    PPSOutput(2, RPB8, OC2);
    SetDCOC2PWM(3825);
    
    mPORTBSetPinsDigitalOut(BIT_5);
    mPORTBSetBits (BIT_5);

    count = 0;
    while (count < 65500000) {
        count ++;
    }
    PPSOutput(2, RPB5, OC2);
    PPSOutput(2, RPB8, NULL);
    
    mPORTBSetPinsDigitalOut(BIT_8);
    mPORTBSetBits(BIT_8);*/
}

static PT_THREAD(protothread_servo(struct pt *pt)) {
    PT_BEGIN(pt);
    static int position = 0;
    static int rotation_count = 0;
    while(1)
    {
        PT_YIELD_TIME_msec(stepper_thread_yield_time);
               
        if (rotation_count < rotation_amount && servo_enabled){
            position = position + rotation_direction;
            if (position > 3) { position = 0; }
            if (position < 0) { position = 3; }
            rotation_count++;
        } else {
            rotation_count = 0;
            servo_enabled = 0;
        }
        
        if (servo_enabled == 0){
            STEPPER_OUT1_CLEAR;
            STEPPER_OUT2_CLEAR;
            STEPPER_OUT3_CLEAR;
            STEPPER_OUT4_CLEAR;
        }
        else if (position == 0)
        {
            STEPPER_OUT1_CLEAR;
            STEPPER_OUT2_SET;
            
            STEPPER_OUT3_CLEAR;
            STEPPER_OUT4_SET;
        }
        else if (position == 1)
        {
            STEPPER_OUT1_SET;
            STEPPER_OUT2_CLEAR;
            
            STEPPER_OUT3_CLEAR;
            STEPPER_OUT4_SET;
        }
        else if (position == 2)
        {
            STEPPER_OUT1_SET;
            STEPPER_OUT2_CLEAR;
            
            STEPPER_OUT3_SET;
            STEPPER_OUT4_CLEAR;
        }
        else if (position == 3)
        {
            STEPPER_OUT1_CLEAR;
            STEPPER_OUT2_SET;
            
            STEPPER_OUT3_SET;
            STEPPER_OUT4_CLEAR;
        }
    } // END WHILE(1)
    PT_END(pt);
} // servo thread

// === Color Thread =================================================
// draw 3 color patches for R,G,B from a random number

static PT_THREAD(protothread_color(struct pt *pt)) {
    PT_BEGIN(pt);
    static int color;
    while (1) {
        // yield time 1 second
        PT_YIELD_TIME_msec(50);

        // draw sys_time
        tft_fillRoundRect(1, 10, 240, 40, 1, ILI9340_BLACK); // x,y,w,h,radius,color
        tft_setCursor(5, 10);
        tft_setTextColor(ILI9340_YELLOW);
        tft_setTextSize(1);

        sprintf(buffer, "System time: %d \n Testing: %s \n Rotation amount: %d", sys_time_seconds, PT_term_buffer, rotation_amount);
        tft_writeString(buffer);

        // choose a random color
        color = rand() & 0xffff;

        // draw color string
        tft_fillRoundRect(0, 50, 150, 14, 1, ILI9340_BLACK); // x,y,w,h,radius,color
        tft_setCursor(0, 50);
        tft_setTextColor(ILI9340_WHITE);
        tft_setTextSize(1);
        sprintf(buffer, " %04x  %04x  %04x  %04x", color & 0x1f, color & 0x7e0, color & 0xf800, color);
        tft_writeString(buffer);

        // draw the actual color patches
        tft_fillRoundRect(5, 70, 30, 30, 1, color & 0x1f); // x,y,w,h,radius,blues
        tft_fillRoundRect(40, 70, 30, 30, 1, color & 0x7e0); // x,y,w,h,radius,greens
        tft_fillRoundRect(75, 70, 30, 30, 1, color & 0xf800); // x,y,w,h,radius,reds
        // now draw the RGB mixed color
        tft_fillRoundRect(110, 70, 30, 30, 1, color); // x,y,w,h,radius,mix color
        // NEVER exit while
    } // END WHILE(1)
    PT_END(pt);
} // color thread

// === Animation Thread =============================================
// move a disk

static PT_THREAD(protothread_anim(struct pt *pt)) {
    PT_BEGIN(pt);
    static int xc = 10, yc = 150, vxc = 2, vyc = 0;
    while (1) {
        // yield time 1 second
        PT_YIELD_TIME_msec(32);

        // erase disk
        tft_fillCircle(xc, yc, 4, ILI9340_BLACK); //x, y, radius, color
        // compute new position
        xc = xc + vxc;
        if (xc < 5 || xc > 235) vxc = -vxc;
        //  draw disk
        tft_fillCircle(xc, yc, 4, ILI9340_GREEN); //x, y, radius, color
        // NEVER exit while
    } // END WHILE(1)
    PT_END(pt);
} // animation thread

// === Input Thread ======================================================
// The serial interface
static char cmd[16];
static int value;

static PT_THREAD(protothread_serial(struct pt *pt)) {
    PT_BEGIN(pt);
    while (1) {
        // send the prompt via DMA to serial
        sprintf(PT_send_buffer, "cmd>");
        // by spawning a print thread
        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));

        //spawn a thread to handle terminal input
        // the input thread waits for input
        // -- BUT does NOT block other threads
        // string is returned in "PT_term_buffer"
        PT_SPAWN(pt, &pt_input, PT_GetSerialBuffer(&pt_input));
        // returns when the thead dies
        // in this case, when <enter> is pushed
        // now parse the string
        sscanf(PT_term_buffer, "%s %d", cmd, &value);

        if (strcmp(cmd, "RST") == 0){
            motor_speed = 30;
            move_robot(move_stop, motor_speed);
            
            stepper_movement_time = 0;
            stepper_thread_yield_time = DEFAULT_STEPPER_THREAD_YIELD_TIME;
            set_stepper_position(180);
        }

        if (strcmp(cmd, "ASPD") == 0){
            stepper_movement_time = value;
        }
        
        if (strcmp(cmd, "ANG") == 0){ // works
            set_stepper_position(value);
        }
        
        if (strcmp(cmd, "SPD") == 0){ // works
            motor_speed = value;
        }
        
        if (strcmp(cmd, "UP") == 0){ // works
            move_robot(move_straight, motor_speed);
        }
        
        if (strcmp(cmd, "DOWN") == 0){ // works
            move_robot(move_back, motor_speed);
        }
        
        if (strcmp(cmd, "LEFT") == 0){ // works
            move_robot(move_left, motor_speed);
        }
        
        if (strcmp(cmd, "RIGHT") == 0){ // works
            move_robot(move_right, motor_speed);
        }
        
        if (strcmp(cmd, "RCCW") == 0){ // NOT USED
            robot_turning = 1;
            stop_time = sys_time_seconds + ROBOT_TURN_TIME;
            move_robot(move_left, ROBOT_TURN_SPEED);
        }
        
        if (strcmp(cmd, "RCW") == 0){ // NOT USED
            robot_turning = 1;
            stop_time = sys_time_seconds + ROBOT_TURN_TIME;
            move_robot(move_right, ROBOT_TURN_SPEED);
        }
        
        if (strcmp(cmd, "STP") == 0){ // works
            move_robot(move_stop, motor_speed);
        }
        
        if (cmd[0] == 'z') printf("%d\t%d\n\r", PT_GET_TIME(), rotation_amount);
        // never exit while
    } // END WHILE(1)
    PT_END(pt);
} // serial thread

// === Main  ======================================================

void main(void) {
    SYSTEMConfigPerformance(PBCLK);
    SYSTEMConfig(SYS_FREQ, SYS_CFG_WAIT_STATES | SYS_CFG_PCACHE);

    ANSELA = 0;
    ANSELB = 0;
    CM1CON = 0;
    CM2CON = 0;

    // === init the USART i/o pins =========
    PPSInput(2, U2RX, RPB11); //Assign U2RX to pin RPB11 -- Physical pin 22 on 28 PDIP
    PPSOutput(4, RPB10, U2TX); //Assign U2TX to pin RPB10 -- Physical pin 21 on 28 PDIP

    // === init the uart2 ===================
    UARTConfigure(UART2, UART_ENABLE_PINS_TX_RX_ONLY);
    UARTSetLineControl(UART2, UART_DATA_SIZE_8_BITS | UART_PARITY_NONE | UART_STOP_BITS_1);
    UARTSetDataRate(UART2, PB_FREQ, BAUDRATE);
    UARTEnable(UART2, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX | UART_TX));
    printf("\n\r..protothreads start..\n\r");

    // === set up DMA for UART output =========
    // configure the channel and enable end-on-match
    DmaChnOpen(DMA_CHANNEL1, DMA_CHN_PRI2, DMA_OPEN_MATCH);
    // trigger a byte everytime the UART is empty
    DmaChnSetEventControl(DMA_CHANNEL1, DMA_EV_START_IRQ_EN | DMA_EV_MATCH_EN | DMA_EV_START_IRQ(_UART2_TX_IRQ));
    // source and destination
    DmaChnSetTxfer(DMA_CHANNEL1, PT_send_buffer + 1, (void*) &U2TXREG, max_chars, 1, 1);
    // signal when done
    DmaChnSetEvEnableFlags(DMA_CHANNEL1, DMA_EV_BLOCK_DONE);
    // set null as ending character (of a string)
    DmaChnSetMatchPattern(DMA_CHANNEL1, 0x00);

    // === config threads ==========
    // turns OFF UART support and debugger pin
    PT_setup();

    // === setup system wide interrupts  ========
    INTEnableSystemMultiVectoredInt();

    // init the threads
    PT_INIT(&pt_servo);
    PT_INIT(&pt_color);
    PT_INIT(&pt_anim);
    PT_INIT(&pt_serial);
    PT_INIT(&pt_timekeeper);

    // init the display
    tft_init_hw();
    tft_begin();
    tft_fillScreen(ILI9340_BLACK);
    //240x320 vertical display
    tft_setRotation(0); // Use tft_setRotation(1) for 320x240

    // seed random color
    srand(1);

    // PWM Period to use (total ticks)
    int generate_period = 10000;
    
    // Default configurations for right wheel (default backward)
    OpenTimer3(T3_ON | T3_SOURCE_INT | T3_PS_1_1, generate_period);
    OpenOC1(OC_ON | OC_TIMER3_SRC | OC_PWM_FAULT_PIN_ENABLE, generate_period >> 1, generate_period >> 2); //OC_PWM_FAULT_PIN_ENABLE
    PPSOutput(1, RPA0, OC1);
    SetDCOC1PWM(3825);  
    mPORTBSetPinsDigitalOut(BIT_7);
    mPORTBClearBits (BIT_7);
    
        
    // Default configurations for left wheel (default forward)
    OpenOC2(OC_ON | OC_TIMER3_SRC | OC_PWM_FAULT_PIN_ENABLE, generate_period >> 1, generate_period >> 2); //OC_PWM_FAULT_PIN_ENABLE
    PPSOutput(2, RPB8, OC2);
    SetDCOC2PWM(3825);
    mPORTBSetPinsDigitalOut(BIT_5);
    mPORTBClearBits (BIT_5);
    
    // Set up stepper ports
    mPORTBSetPinsDigitalOut(BIT_9);  // STEPPER_OUT1
    mPORTBSetPinsDigitalOut(BIT_3);  // STEPPER_OUT2
    mPORTBSetPinsDigitalOut(BIT_13); // STEPPER_OUT3
    mPORTBSetPinsDigitalOut(BIT_15); // STEPPER_OUT4

    STEPPER_OUT1_CLEAR;
    STEPPER_OUT2_CLEAR;
    STEPPER_OUT3_CLEAR;
    STEPPER_OUT4_CLEAR;
    
    // Set up motor ports
    mPORTASetPinsDigitalOut(BIT_2);
    //mPORTASetPinsDigitalOut(BIT_3);
    //mPORTASetPinsDigitalOut(BIT_0);    

    mPORTAClearBits(BIT_2);
    //mPORTAClearBits(BIT_3);
    //mPORTASetBits(BIT_0);
    
    //set_stepper_position(90);
    
    // round-robin scheduler for threads
    while (1) {
        PT_SCHEDULE(protothread_servo(&pt_servo));
        //PT_SCHEDULE(protothread_color(&pt_color));
        //PT_SCHEDULE(protothread_anim(&pt_anim));
        PT_SCHEDULE(protothread_serial(&pt_serial));
        PT_SCHEDULE(protothread_timekeeper(&pt_timekeeper));
    }
} // main

// === end  ======================================================